import SchemaManager from './SchemaManager';
export declare enum DataType {
    Int = "int",
    Float = "float",
    Double = "double",
    Long = "long",
    String = "string",
    Boolean = "bool",
    Date = "date",
    Enum = "enum"
}
export interface ColumnOptions {
    type?: DataType;
    length?: number;
    field?: string;
    required?: boolean;
    primaryKey?: boolean;
    generatedKey?: boolean;
    refName?: string;
    refKey?: boolean;
    middle?: Schema;
    middleKey?: string;
    defaultValue?: string;
    initialValue?: string;
    defaultFunction?: Function;
    initialFunction?: Function;
}
export declare class Column {
    name: string;
    readonly schema: Schema;
    readonly options: ColumnOptions;
    constructor(schema: Schema, name: string, options: ColumnOptions);
    readonly primaryKey: boolean;
    readonly generatedKey: boolean;
    readonly ref: Schema | undefined;
    readonly refKey: boolean;
    readonly field: string;
    readonly type: DataType;
    readonly length: number | undefined;
    readonly required: boolean;
    private static getDefaultLength;
    toJSON(): {
        type?: DataType | undefined;
        length?: number | undefined;
        field?: string | undefined;
        required?: boolean | undefined;
        primaryKey?: boolean | undefined;
        generatedKey?: boolean | undefined;
        refName?: string | undefined;
        refKey?: boolean | undefined;
        middle?: Schema | undefined;
        middleKey?: string | undefined;
        defaultValue?: string | undefined;
        initialValue?: string | undefined;
        defaultFunction?: Function | undefined;
        initialFunction?: Function | undefined;
        name: string;
    };
}
export declare enum AssociationType {
    BelongsTo = "belongsTo",
    HasOne = "hasOne",
    HasMany = "hasMany",
    BelongsToMany = "belongsToMany"
}
export interface AssociationOptions {
    refName?: string;
    type?: AssociationType;
    refKey?: string;
    required?: boolean;
    middleName?: string;
    middleKey?: string;
    cascade?: 'none' | 'delete';
}
export declare class Association {
    private schemaManager;
    name: string;
    readonly current: Schema;
    readonly options: AssociationOptions;
    constructor(schemaManager: SchemaManager, name: string, current: Schema, options: AssociationOptions);
    readonly ref: Schema;
    readonly refName: string;
    readonly type: AssociationType;
    readonly required: boolean;
    readonly refKey: string;
    readonly middle: Schema;
    readonly middleName: string;
    readonly middleKey: string;
    private static getDefaultRefKey;
    toJSON(): {
        refName?: string | undefined;
        type?: AssociationType | undefined;
        refKey?: string | undefined;
        required?: boolean | undefined;
        middleName?: string | undefined;
        middleKey?: string | undefined;
        cascade?: "none" | "delete" | undefined;
        name: string;
    };
}
export interface SchemaOptions {
    table?: string;
}
export default class Schema {
    name: string;
    private idColumn?;
    readonly columns: Column[];
    readonly options: SchemaOptions;
    readonly associations: Association[];
    readonly schemaManager: SchemaManager;
    constructor(schemaManager: SchemaManager, name: string, options?: SchemaOptions);
    readonly table: string;
    private createColumn;
    /**
     * 添加column
     * @param name
     * @param options
     */
    addColumn(name: string, options: ColumnOptions): Schema;
    /**
     * 修改column
     * @param oldName
     * @param newName
     * @param options
     */
    updateColumn(oldName: string, newName: string, options: ColumnOptions): Schema;
    /**
     * 删除column
     * @param name
     */
    deleteColumn(name: string): void;
    /**
     * 判断是否有column
     * @param name
     */
    hasColumn(name: string): boolean;
    /**
     * 判断是否有association
     * @param name
     */
    hasAssociation(name: string): boolean;
    /**
     * 获取column
     * @param name
     */
    getColumn(name: string): Column | undefined;
    /**
     * 获取主键
     */
    getIdColumn(): Column | undefined;
    /**
     * 获取association
     * @param name
     */
    getAssociation(name: string): Association | undefined;
    /**
     * 添加关联column
     * @param name
     * @param type
     * @param length
     * @param refName
     */
    private addRefColumn;
    private createAssociation;
    /**
     * 添加association
     * @param name
     * @param options
     */
    addAssociation(name: string, options: AssociationOptions): Schema;
    /**
     * 删除association
     * @param name
     */
    deleteAssociation(name: string): void;
    /**
     * 修改association
     * @param oldName
     * @param name
     * @param options
     */
    updateAssociation(oldName: string, name: string, options: AssociationOptions): void;
    /**
     * 获取外键
     * @param name
     */
    getAssociationRefColumn(name: string): Column;
    toJSON(): {
        name: string;
        table: string | undefined;
        columns: Column[];
        associations: Association[];
    };
}
