"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var DataType;
(function (DataType) {
    DataType["Int"] = "int";
    DataType["Float"] = "float";
    DataType["Double"] = "double";
    DataType["Long"] = "long";
    DataType["String"] = "string";
    DataType["Boolean"] = "bool";
    DataType["Date"] = "date";
    DataType["Enum"] = "enum";
})(DataType = exports.DataType || (exports.DataType = {}));
class Column {
    constructor(schema, name, options) {
        this.schema = schema;
        this.name = name;
        this.options = options;
        if (this.options.defaultValue != undefined) {
            this.options.defaultFunction = new Function(`return ${this.options.defaultValue}`);
        }
        if (this.options.initialValue != undefined) {
            this.options.initialFunction = new Function(`return ${this.options.initialValue}`);
        }
    }
    get primaryKey() {
        return this.options.primaryKey || false;
    }
    get generatedKey() {
        return this.options.generatedKey || false;
    }
    get ref() {
        return this.schema.schemaManager.getSchema(this.options.refName);
    }
    get refKey() {
        return this.options.refKey == true;
    }
    get field() {
        return this.options.field || this.name;
    }
    get type() {
        return this.options.type;
    }
    get length() {
        return this.options.length || Column.getDefaultLength(this.options.type);
    }
    get required() {
        return this.options.required || false;
    }
    // 获取默认长度
    static getDefaultLength(type) {
        switch (type) {
            case DataType.String:
                return 256;
            default:
                return undefined;
        }
    }
    toJSON() {
        return Object.assign({ name: this.name }, this.options);
    }
}
exports.Column = Column;
var AssociationType;
(function (AssociationType) {
    // 属于
    AssociationType["BelongsTo"] = "belongsTo";
    // 拥有一个
    AssociationType["HasOne"] = "hasOne";
    // 拥有多个
    AssociationType["HasMany"] = "hasMany";
    // 属于多个
    AssociationType["BelongsToMany"] = "belongsToMany";
})(AssociationType = exports.AssociationType || (exports.AssociationType = {}));
class Association {
    constructor(schemaManager, name, current, options) {
        this.schemaManager = schemaManager;
        this.name = name;
        this.current = current;
        this.options = options;
    }
    get ref() {
        return this.schemaManager.getSchema(this.options.refName);
    }
    get refName() {
        return this.options.refName;
    }
    get type() {
        return this.options.type;
    }
    get required() {
        return this.options.required || true;
    }
    get refKey() {
        return this.options.refKey || Association.getDefaultRefKey(this.name, this.current, this.type);
    }
    get middle() {
        return this.schemaManager.getSchema(this.middleName);
    }
    get middleName() {
        return this.options.middleName;
    }
    get middleKey() {
        // post belongsToMany tag
        // ref tag middle postTag
        // middle key tagId
        return this.options.middleKey || `${this.ref.name}Id`;
    }
    static getDefaultRefKey(name, current, type) {
        switch (type) {
            case AssociationType.BelongsTo:
                // user belongsTo role  user.roleId
                return `${name}Id`;
            case AssociationType.HasOne:
                // user hasOne info  info.userId
                return `${name}Id`;
            case AssociationType.HasMany:
                // role hasMany users  user.roleId
                return `${current.name}Id`;
            case AssociationType.BelongsToMany:
                // post belongsToMany tags  middle.postId
                return `${current.name}Id`;
        }
    }
    toJSON() {
        return Object.assign({ name: this.name }, this.options);
    }
}
exports.Association = Association;
class Schema {
    constructor(schemaManager, name, options = {}) {
        this.columns = [];
        this.associations = [];
        this.schemaManager = schemaManager;
        this.name = name;
        this.options = options;
    }
    get table() {
        return this.options.table || this.name;
    }
    createColumn(name, options) {
        const column = new Column(this, name, options);
        if (column.primaryKey)
            this.idColumn = column;
        return column;
    }
    /**
     * 添加column
     * @param name
     * @param options
     */
    addColumn(name, options) {
        const column = this.createColumn(name, options);
        this.columns.push(column);
        return this;
    }
    /**
     * 修改column
     * @param oldName
     * @param newName
     * @param options
     */
    updateColumn(oldName, newName, options) {
        const index = this.columns.findIndex(column => column.name == oldName);
        this.columns[index] = this.createColumn(newName, options);
        return this;
    }
    /**
     * 删除column
     * @param name
     */
    deleteColumn(name) {
        const index = this.columns.findIndex(column => column.name == name);
        if (index >= 0)
            this.columns.splice(index, 1);
    }
    /**
     * 判断是否有column
     * @param name
     */
    hasColumn(name) {
        return this.columns.find(column => column.name == name) != undefined;
    }
    /**
     * 判断是否有association
     * @param name
     */
    hasAssociation(name) {
        return this.associations.find(association => association.name == name) != undefined;
    }
    /**
     * 获取column
     * @param name
     */
    getColumn(name) {
        return this.columns.find(column => column.name == name);
    }
    /**
     * 获取主键
     */
    getIdColumn() {
        return this.idColumn;
    }
    /**
     * 获取association
     * @param name
     */
    getAssociation(name) {
        return this.associations.find(association => association.name == name);
    }
    /**
     * 添加关联column
     * @param name
     * @param type
     * @param length
     * @param refName
     */
    addRefColumn(name, type, length, refName) {
        if (this.hasColumn(name))
            return;
        this.addColumn(name, { type, refKey: true, refName, length });
    }
    createAssociation(name, options) {
        const association = new Association(this.schemaManager, name, this, options);
        switch (association.type) {
            case AssociationType.HasOne:
            case AssociationType.HasMany:
                // 插入外键到ref
                const id = this.getIdColumn();
                association.ref.addRefColumn(association.refKey, id.type, id.length, association.refName);
                break;
            case AssociationType.BelongsTo:
                // 插入外键
                const refId = association.ref.getIdColumn();
                this.addRefColumn(association.refKey, refId.type, refId.length, association.refName);
                break;
        }
        return association;
    }
    /**
     * 添加association
     * @param name
     * @param options
     */
    addAssociation(name, options) {
        const association = this.createAssociation(name, options);
        this.associations.push(association);
        return this;
    }
    /**
     * 删除association
     * @param name
     */
    deleteAssociation(name) {
        const index = this.associations.findIndex(association => association.name == name);
        if (index >= 0)
            this.associations.splice(index, 1);
    }
    /**
     * 修改association
     * @param oldName
     * @param name
     * @param options
     */
    updateAssociation(oldName, name, options) {
        const association = this.createAssociation(name, options);
        const index = this.associations.findIndex(association => association.name == oldName);
        this.associations[index] = association;
    }
    /**
     * 获取外键
     * @param name
     */
    getAssociationRefColumn(name) {
        const association = this.getAssociation(name);
        switch (association.type) {
            case AssociationType.HasOne:
            case AssociationType.HasMany:
                return association.ref.getColumn(association.refKey);
            case AssociationType.BelongsTo:
                return this.getColumn(association.refKey);
        }
        throw new Error('');
    }
    toJSON() {
        return {
            name: this.name,
            table: this.options.table,
            columns: this.columns,
            associations: this.associations
        };
    }
}
exports.default = Schema;
