"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const SqlNode_1 = require("../sql/SqlNode");
class BaseDialect {
    gen(node) {
        if (node instanceof SqlNode_1.SqlQuery)
            return this.genQuery(node);
        if (node instanceof SqlNode_1.SqlInsert)
            return this.genInsert(node);
        if (node instanceof SqlNode_1.SqlUpdate)
            return this.genUpdate(node);
        if (node instanceof SqlNode_1.SqlDelete)
            return this.genDelete(node);
        throw new Error('');
    }
    genQuery(sqlQuery) {
        return 'select ' +
            this.genSelect(sqlQuery.select) +
            this.genFrom(sqlQuery.from) +
            this.genJoins(sqlQuery.joins) +
            this.genWhere(sqlQuery.where) +
            this.genOrders(sqlQuery.orders) +
            this.genPage(sqlQuery.page);
    }
    genSelect(select) {
        return select.map(column => this.genSelectColumn(column)).join(', ');
    }
    genSelectColumn(column) {
        if (column instanceof SqlNode_1.SqlCountColumn)
            return `count(${column.name})`;
        return column.table != undefined ? `${column.table}.${column.name} as ${column.table}_${column.name}` : column.name;
    }
    genFrom(from) {
        if (from instanceof SqlNode_1.SqlTableForm)
            return this.genFromSqlTable(from.table);
        if (from instanceof SqlNode_1.SqlInnerFrom) {
            const innerQuery = from.query;
            // 只支持一层嵌套
            return ` from (${this.genQuery(innerQuery)}) as ${innerQuery.from.table.name}`;
        }
        throw new Error('');
    }
    genFromSqlTable(sqlTable) {
        if (sqlTable.alias)
            return ` from ${sqlTable.name} as ${sqlTable.alias}`;
        return ` from ${sqlTable.name}`;
    }
    genJoins(joins) {
        if (joins.length == 0)
            return '';
        return joins.map(join => this.genJoin(join)).join(' ');
    }
    genJoin(join) {
        const joinSql = join.type == SqlNode_1.SqlJoinType.Inner ? " inner join " : " left join ";
        const expSql = this.genExp(join.on);
        return joinSql + join.table + " as " + join.alias + " on " + expSql;
    }
    genExp(exp) {
        if (exp instanceof SqlNode_1.SqlLogicExp) {
            const left = this.genExp(exp.left);
            const op = this.genLogicOp(exp.op);
            const right = this.genExp(exp.right);
            return `${left} ${op} ${right}`;
        }
        else if (exp instanceof SqlNode_1.SqlColumnExp) {
            const left = this.genColumn(exp.left);
            const op = this.genCompareOp(exp.op);
            let right = '';
            if (exp.right instanceof SqlNode_1.SqlColumn) {
                right = this.genColumn(exp.right);
            }
            else if (exp.right instanceof SqlNode_1.SqlValue) {
                right = this.genValue(exp.right);
            }
            else if (exp.right instanceof SqlNode_1.SqlParam) {
                right = this.genParam(exp.right);
            }
            if (op == '=' && right == 'null') {
                return `${left} is null`;
            }
            if (op == '<>' && right == 'null') {
                return `${left} is not null`;
            }
            return `${left} ${op} ${right}`;
        }
        else if (exp instanceof SqlNode_1.SqlNestExp) {
            return `(${this.genExp(exp.exp)})`;
        }
        throw new Error('');
    }
    genColumn(column) {
        return column.table != undefined ? `${column.table}.${column.name}` : column.name;
    }
    genParam(param) {
        return `$${param.name}`;
    }
    genValue(right) {
        if (typeof right.value == 'string')
            return JSON.stringify(right.value);
        return `${right.value}`;
    }
    genCompareOp(op) {
        switch (op) {
            case SqlNode_1.SqlCompareOp.Eq:
                return '=';
            case SqlNode_1.SqlCompareOp.Ge:
                return '>=';
            case SqlNode_1.SqlCompareOp.Gt:
                return '>';
            case SqlNode_1.SqlCompareOp.Le:
                return '<=';
            case SqlNode_1.SqlCompareOp.Lt:
                return '<';
            case SqlNode_1.SqlCompareOp.Ne:
                return '<>';
            case SqlNode_1.SqlCompareOp.Like:
                return 'like';
        }
    }
    genLogicOp(op) {
        return op == SqlNode_1.SqlLogicOp.And ? '&&' : '||';
    }
    genWhere(where) {
        if (where.length == 0)
            return '';
        return ' where ' + where.map(exp => this.genExp(exp)).join(' and ');
    }
    genOrders(orders) {
        if (orders.length == 0)
            return '';
        return ' order by ' + orders.map(order => order.columns.map(column => column.name).join(', ') + ' ' + order.sort).join(' ');
    }
    genPage(page) {
        if (page.limit == undefined)
            return '';
        return page.offset != undefined ? ` limit ${page.offset}, ${page.limit}` : ` limit ${page.limit}`;
    }
    genInsert(insert) {
        return `insert into ${insert.table} ` +
            `(${insert.columns.map(column => column.name).join(', ')}) ` +
            `values (${insert.params.map(param => `$${[param.name]}`).join(', ')})`;
    }
    genUpdate(update) {
        return `update ${update.table} ` +
            `set ${update.sets.map(set => `${set.name} = $${set.param}`).join(', ')} ` +
            `where ${this.genExp(update.where)}`;
    }
    genDelete(node) {
        return `delete from ${node.table} where ${this.genExp(node.where)}`;
    }
}
exports.default = BaseDialect;
