"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const OrqlNode_1 = require("orql-parser/lib/OrqlNode");
const SqlNode_1 = require("./SqlNode");
const Schema_1 = require("../Schema");
class OrqlToSql {
    constructor(schemaManager, dialect) {
        this.sqlCaches = new Map();
        this.schemaManager = schemaManager;
        this.dialect = dialect;
    }
    toQuery(node, options = {}) {
        let sqlPage = new SqlNode_1.SqlPage(options.offset, options.limit);
        const root = node.item;
        const op = node.op;
        const rootSchema = this.schemaManager.getSchema(root.name);
        const table = rootSchema.table;
        const select = [];
        const joins = [];
        const where = [];
        const orders = [];
        // 根节点exp
        let rootExp = undefined;
        // 根节点order
        const rootOrders = [];
        const queryStack = [{ currentItem: root, currentPath: table, isRoot: true, schema: rootSchema }];
        // 存在数组关联类型
        let hasArrayRef = false;
        while (queryStack.length > 0) {
            const { currentItem, currentPath, isRoot, schema } = queryStack.pop();
            const currentSchema = schema;
            const idColumn = currentSchema.getIdColumn();
            // 是否select id
            let hasSelectId = false;
            // 是否有select
            let hasSelect = false;
            // 是否有all
            let hasAll = false;
            // 忽略项
            const ignores = [];
            if (currentItem.where) {
                if (currentItem.where.exp) {
                    const exp = this.toExp(currentItem.where.exp, currentSchema, currentPath);
                    if (isRoot) {
                        rootExp = exp;
                    }
                    else {
                        where.push(exp);
                    }
                }
                if (currentItem.where.orders) {
                    for (const order of currentItem.where.orders) {
                        // 排序项全部列
                        const columns = order.columns.map(item => {
                            const column = currentSchema.getColumn(item.name);
                            return new SqlNode_1.SqlColumn(column.field || column.name, currentPath);
                        });
                        const sqlOrder = new SqlNode_1.SqlOrder(order.sort, columns);
                        if (isRoot)
                            rootOrders.push(sqlOrder);
                        // 嵌套内外order都要一样
                        orders.push(sqlOrder);
                    }
                }
            }
            if (currentItem.children) {
                for (const childItem of currentItem.children) {
                    // 有select
                    hasSelect = true;
                    if (childItem instanceof OrqlNode_1.OrqlAllItem) {
                        hasAll = true;
                    }
                    else if (childItem instanceof OrqlNode_1.OrqlIgnoreItem) {
                        // remove ignore
                        ignores.push(childItem.name);
                    }
                    else if (currentSchema.hasColumn(childItem.name)) {
                        const column = currentSchema.getColumn(childItem.name);
                        // 有select id
                        if (column.primaryKey)
                            hasSelectId = true;
                        if (op == 'query')
                            select.push(new SqlNode_1.SqlColumn(column.field, currentPath));
                    }
                    else if (currentSchema.hasAssociation(childItem.name)) {
                        const association = currentSchema.getAssociation(childItem.name);
                        const childSchema = association.ref;
                        const childIdColumn = childSchema.getIdColumn();
                        const childPath = currentPath != undefined ? `${currentPath}_${childItem.name}` : childItem.name;
                        // 入栈
                        queryStack.push({ currentItem: childItem, currentPath: childPath, schema: childSchema });
                        const joinType = association.required ? SqlNode_1.SqlJoinType.Inner : SqlNode_1.SqlJoinType.Left;
                        switch (association.type) {
                            case Schema_1.AssociationType.HasMany:
                                if (childItem.children && childItem.children.length > 0)
                                    hasArrayRef = true;
                                // role hasMany user
                                // user.roleId = role.id
                                const hasManyOn = new SqlNode_1.SqlColumnExp(new SqlNode_1.SqlColumn(association.refKey, childPath), SqlNode_1.SqlCompareOp.Eq, new SqlNode_1.SqlColumn(childIdColumn.field, currentPath));
                                joins.push(new SqlNode_1.SqlJoin(childSchema.table, childPath, joinType, hasManyOn));
                                break;
                            case Schema_1.AssociationType.HasOne:
                                // user hasOne info
                                // info.userId = user.id
                                const hasOneOn = new SqlNode_1.SqlColumnExp(new SqlNode_1.SqlColumn(association.refKey, childPath), SqlNode_1.SqlCompareOp.Eq, new SqlNode_1.SqlColumn(childIdColumn.field, currentPath));
                                joins.push(new SqlNode_1.SqlJoin(childSchema.table, childPath, joinType, hasOneOn));
                                break;
                            case Schema_1.AssociationType.BelongsTo:
                                // user belongsTo role
                                // role.id = user.roleId
                                const belongsToOn = new SqlNode_1.SqlColumnExp(new SqlNode_1.SqlColumn(association.ref.getIdColumn().field, childPath), SqlNode_1.SqlCompareOp.Eq, new SqlNode_1.SqlColumn(association.refKey, currentPath));
                                joins.push(new SqlNode_1.SqlJoin(childSchema.table, childPath, joinType, belongsToOn));
                                break;
                            case Schema_1.AssociationType.BelongsToMany:
                                if (childItem.children && childItem.children.length > 0)
                                    hasArrayRef = true;
                                // post belongsToMany tag middle postTag
                                // postTag.postId(refKey) = post.id
                                // postTag.tagId(middleKey) = tag.id
                                const targetSchema = association.current;
                                const middleSchema = association.middle;
                                const refSchema = association.ref;
                                const middlePath = `${currentPath}_${middleSchema.name}`;
                                const leftOn = new SqlNode_1.SqlColumnExp(new SqlNode_1.SqlColumn(association.refKey, middlePath), SqlNode_1.SqlCompareOp.Eq, new SqlNode_1.SqlColumn(childIdColumn.field, currentPath));
                                joins.push(new SqlNode_1.SqlJoin(association.middle.name, middlePath, joinType, leftOn));
                                const rightOn = new SqlNode_1.SqlColumnExp(new SqlNode_1.SqlColumn(targetSchema.getIdColumn().field, childPath), SqlNode_1.SqlCompareOp.Eq, new SqlNode_1.SqlColumn(association.middleKey, middlePath));
                                joins.push(new SqlNode_1.SqlJoin(refSchema.table, childPath, joinType, rightOn));
                                break;
                        }
                    }
                }
            }
            if (hasAll) {
                // add all
                for (const column of currentSchema.columns) {
                    if (!column.refKey && op == 'query' && ignores.indexOf(column.name) < 0) {
                        select.push(new SqlNode_1.SqlColumn(column.field, currentPath));
                    }
                }
            }
            if (!hasSelectId && op == 'query' && hasSelect && idColumn) {
                // 没有select id而且是query而且有select插入id进行mapper
                // 没有id的不加入进来，在mapper把每一个列都当成唯一的
                select.push(new SqlNode_1.SqlColumn(idColumn.field, currentPath));
            }
        }
        let query;
        if (op == 'count') {
            //分页
            select.push(new SqlNode_1.SqlCountColumn(rootSchema.getIdColumn().field, table));
            if (rootExp != null)
                where.unshift(rootExp);
            const from = new SqlNode_1.SqlTableForm(new SqlNode_1.SqlTable(table, table));
            query = new SqlNode_1.SqlQuery(select, from, where, joins, orders, sqlPage);
        }
        else if (hasArrayRef && sqlPage.limit != undefined) {
            //嵌套分页查询
            const innerSelect = [new SqlNode_1.SqlColumn("*")];
            const innerWhere = rootExp != undefined ? [rootExp] : [];
            const innerFrom = new SqlNode_1.SqlTableForm(new SqlNode_1.SqlTable(table));
            const from = new SqlNode_1.SqlInnerFrom(new SqlNode_1.SqlQuery(innerSelect, innerFrom, innerWhere, [], rootOrders, sqlPage));
            query = new SqlNode_1.SqlQuery(select, from, where, joins, orders, new SqlNode_1.SqlPage());
        }
        else if (!hasArrayRef && sqlPage.limit == undefined && !root.isArray) {
            //无分页，单个查询，而且没有数组类型关联查询
            if (rootExp != null)
                where.unshift(rootExp);
            const from = new SqlNode_1.SqlTableForm(new SqlNode_1.SqlTable(table, table));
            sqlPage = new SqlNode_1.SqlPage(undefined, 1);
            query = new SqlNode_1.SqlQuery(select, from, where, joins, orders, sqlPage);
        }
        else {
            if (rootExp != null)
                where.unshift(rootExp);
            const from = new SqlNode_1.SqlTableForm(new SqlNode_1.SqlTable(table, table));
            query = new SqlNode_1.SqlQuery(select, from, where, joins, orders, sqlPage);
        }
        return this.dialect.gen(query);
    }
    toExp(orqlExp, parent, path) {
        if (orqlExp instanceof OrqlNode_1.OrqlLogicExp) {
            const left = this.toExp(orqlExp.left, parent, path);
            const op = orqlExp.op == OrqlNode_1.OrqlLogicOp.And ? SqlNode_1.SqlLogicOp.And : SqlNode_1.SqlLogicOp.Or;
            const right = this.toExp(orqlExp.right, parent, path);
            return new SqlNode_1.SqlLogicExp(left, op, right);
        }
        if (orqlExp instanceof OrqlNode_1.OrqlCompareExp) {
            const leftColumn = parent.getColumn(orqlExp.left.name);
            // FIXME 报错
            const left = new SqlNode_1.SqlColumn(leftColumn.field, path);
            const op = this.toCompareOp(orqlExp.op);
            let right;
            if (orqlExp.right instanceof OrqlNode_1.OrqlColumn) {
                const rightColumn = parent.getColumn(orqlExp.right.name);
                // FIXME 报错
                right = new SqlNode_1.SqlColumn(rightColumn.name, path);
            }
            else if (orqlExp.right instanceof OrqlNode_1.OrqlValue) {
                right = new SqlNode_1.SqlValue(orqlExp.right.value);
            }
            else if (orqlExp.right instanceof OrqlNode_1.OrqlParam) {
                right = new SqlNode_1.SqlParam(orqlExp.right.name);
            }
            return new SqlNode_1.SqlColumnExp(left, op, right);
        }
        if (orqlExp instanceof OrqlNode_1.OrqlNestExp) {
            return new SqlNode_1.SqlNestExp(this.toExp(orqlExp.exp, parent, path));
        }
        throw new Error('');
    }
    toLogicOp(orqlLogicOp) {
        switch (orqlLogicOp) {
            case OrqlNode_1.OrqlLogicOp.And:
                return SqlNode_1.SqlLogicOp.And;
            case OrqlNode_1.OrqlLogicOp.Or:
                return SqlNode_1.SqlLogicOp.Or;
        }
    }
    toCompareOp(orqlCompareOp) {
        switch (orqlCompareOp) {
            case OrqlNode_1.OrqlCompareOp.Ge:
                return SqlNode_1.SqlCompareOp.Ge;
            case OrqlNode_1.OrqlCompareOp.Gt:
                return SqlNode_1.SqlCompareOp.Gt;
            case OrqlNode_1.OrqlCompareOp.Le:
                return SqlNode_1.SqlCompareOp.Le;
            case OrqlNode_1.OrqlCompareOp.Lt:
                return SqlNode_1.SqlCompareOp.Lt;
            case OrqlNode_1.OrqlCompareOp.Eq:
                return SqlNode_1.SqlCompareOp.Eq;
            case OrqlNode_1.OrqlCompareOp.Ne:
                return SqlNode_1.SqlCompareOp.Ne;
            case OrqlNode_1.OrqlCompareOp.Like:
                return SqlNode_1.SqlCompareOp.Like;
        }
    }
    toAdd(node, params) {
        // if (this.sqlCaches.has(node)) return this.dialect.gen(this.sqlCaches.get(node)!);
        const root = node.item;
        const columns = [];
        const sqlParams = [];
        if (root.children == undefined) {
            throw new Error(`${root.name} children not exist`);
        }
        const schema = this.schemaManager.getSchema(root.name);
        let hasAll = false;
        const ignores = [];
        for (const item of root.children) {
            if (item instanceof OrqlNode_1.OrqlAllItem) {
                hasAll = true;
            }
            else if (item instanceof OrqlNode_1.OrqlIgnoreItem) {
                ignores.push(item.name);
            }
            else if (schema.hasColumn(item.name)) {
                const column = schema.getColumn(item.name);
                columns.push(new SqlNode_1.SqlColumn(column.field));
                if (column.options.initialValue && params) {
                    params[column.name] = column.options.initialFunction();
                }
                else if (column.options.defaultValue && params && params[column] == undefined) {
                    params[column.name] = column.options.defaultFunction();
                }
                sqlParams.push(new SqlNode_1.SqlParam(column.name));
            }
            else if (schema.hasAssociation(item.name)) {
                const association = schema.getAssociation(item.name);
                switch (association.type) {
                    case Schema_1.AssociationType.BelongsTo:
                        const refIdKey = `${association.name}.${association.ref.getIdColumn().name}`;
                        columns.push(new SqlNode_1.SqlColumn(association.refKey));
                        sqlParams.push(new SqlNode_1.SqlParam(refIdKey));
                        break;
                }
            }
        }
        if (hasAll) {
            for (const column of schema.columns) {
                // 非外键而且不ignore
                if (!column.refKey && ignores.indexOf(column.name) < 0) {
                    columns.push(new SqlNode_1.SqlColumn(column.field));
                    if (column.options.initialValue && params) {
                        params[column.name] = column.options.initialFunction();
                    }
                    else if (column.options.defaultValue && params && params[column] == undefined) {
                        params[column.name] = column.options.defaultFunction();
                    }
                    sqlParams.push(new SqlNode_1.SqlParam(column.name));
                }
            }
        }
        const insert = new SqlNode_1.SqlInsert(schema.table, columns, sqlParams);
        this.sqlCaches.set(node, insert);
        return this.dialect.gen(insert);
    }
    toUpdate(node) {
        if (this.sqlCaches.has(node))
            return this.dialect.gen(this.sqlCaches.get(node));
        const root = node.item;
        const schema = this.schemaManager.getSchema(root.name);
        if (root.where == undefined || root.where.exp == undefined)
            throw new Error('update exp null');
        const exp = this.toExp(root.where.exp, schema, schema.table);
        const sets = [];
        if (root.children == undefined) {
            throw new Error(`update not children`);
        }
        let hasAll = false;
        const ignores = [];
        for (const item of root.children) {
            if (item instanceof OrqlNode_1.OrqlAllItem) {
                hasAll = true;
            }
            else if (item instanceof OrqlNode_1.OrqlIgnoreItem) {
                ignores.push(item.name);
            }
            else if (schema.hasColumn(item.name)) {
                const column = schema.getColumn(item.name);
                sets.push(new SqlNode_1.SqlSet(column.field, column.name));
            }
            else if (schema.hasAssociation(item.name)) {
                const association = schema.getAssociation(item.name);
                switch (association.type) {
                    case Schema_1.AssociationType.BelongsTo:
                        sets.push(new SqlNode_1.SqlSet(association.refKey, `${association.name}.${association.ref.getIdColumn().name}`));
                        break;
                }
            }
        }
        if (hasAll) {
            for (const column of schema.columns) {
                if (!column.refKey && ignores.indexOf(column.name) < 0) {
                    sets.push(new SqlNode_1.SqlSet(column.field, column.name));
                }
            }
        }
        const update = new SqlNode_1.SqlUpdate(schema.table, exp, sets);
        this.sqlCaches.set(node, update);
        return this.dialect.gen(update);
    }
    toDelete(node) {
        if (this.sqlCaches.has(node))
            return this.dialect.gen(this.sqlCaches.get(node));
        const root = node.item;
        if (root.where == undefined || root.where.exp == undefined)
            throw new Error('delete exp null');
        const schema = this.schemaManager.getSchema(root.name);
        if (schema == undefined) {
            throw new Error(`schema ${root.name} not exist`);
        }
        const exp = this.toExp(root.where.exp, schema, schema.table);
        const sqlDelete = new SqlNode_1.SqlDelete(schema.table, exp);
        this.sqlCaches.set(node, sqlDelete);
        return this.dialect.gen(sqlDelete);
    }
}
exports.default = OrqlToSql;
